from enum import Enum
from typing import Generic, List, Optional, TypeVar

import strawberry
from django.db.models.sql.query import get_field_names_from_opts
from strawberry import UNSET
from strawberry.arguments import StrawberryArgument

from . import utils
from .arguments import argument


T = TypeVar("T")


@strawberry.input
class DjangoModelFilterInput:
    pk: strawberry.ID


@strawberry.input
class FilterLookup(Generic[T]):
    exact: Optional[T] = UNSET
    i_exact: Optional[T] = UNSET
    contains: Optional[T] = UNSET
    i_contains: Optional[T] = UNSET
    in_list: Optional[List[T]] = UNSET
    gt: Optional[T] = UNSET
    gte: Optional[T] = UNSET
    lt: Optional[T] = UNSET
    lte: Optional[T] = UNSET
    starts_with: Optional[T] = UNSET
    i_starts_with: Optional[T] = UNSET
    ends_with: Optional[T] = UNSET
    i_ends_with: Optional[T] = UNSET
    range: Optional[List[T]] = UNSET
    is_null: Optional[bool] = UNSET
    regex: Optional[str] = UNSET
    i_regex: Optional[str] = UNSET


lookup_name_conversion_map = {
    "i_exact": "iexact",
    "i_contains": "icontains",
    "in_list": "in",
    "starts_with": "startswith",
    "i_starts_with": "istartswith",
    "ends_with": "endswith",
    "i_ends_with": "iendswith",
    "is_null": "isnull",
    "i_regex": "iregex",
}


def filter(model, *, name=None, lookups=False):
    def wrapper(cls):
        is_filter = lookups and "lookups" or True
        from .type import process_type

        type_ = process_type(
            cls, model, is_input=True, partial=True, is_filter=is_filter
        )
        return type_

    return wrapper


def filter_deprecated(model, *, name=None, lookups=False):
    utils.deprecated(
        "'strawberry_django.filter' is deprecated,"
        " use 'strawberry_django.filters.filter' instead",
        stacklevel=2,
    )
    return filter(model, name=name, lookups=lookups)


def build_filter_kwargs(filters):
    filter_kwargs = {}
    filter_methods = []
    django_model = utils.get_django_model(filters)
    for field in utils.fields(filters):
        field_name = field.name
        field_value = getattr(filters, field_name)

        if field_value is UNSET:
            continue

        if isinstance(field_value, Enum):
            field_value = field_value.value

        filter_method = getattr(filters, f"filter_{field_name}", None)
        if filter_method:
            filter_methods.append(filter_method)
            continue

        if django_model:
            if field_name not in get_field_names_from_opts(django_model._meta):
                continue

        if field_name in lookup_name_conversion_map:
            field_name = lookup_name_conversion_map[field_name]
        if utils.is_strawberry_type(field_value):
            (
                subfield_filter_kwargs,
                subfield_filter_methods,
            ) = build_filter_kwargs(field_value)
            for (
                subfield_name,
                subfield_value,
            ) in subfield_filter_kwargs.items():
                if isinstance(subfield_value, Enum):
                    subfield_value = subfield_value.value
                filter_kwargs[f"{field_name}__{subfield_name}"] = subfield_value
            filter_methods.extend(subfield_filter_methods)
        else:
            filter_kwargs[field_name] = field_value

    return filter_kwargs, filter_methods


def apply(filters, queryset, pk=UNSET):
    if pk is not UNSET:
        queryset = queryset.filter(pk=pk)

    if (
        filters is UNSET
        or filters is None
        or not hasattr(filters, "_django_type")
        or not filters._django_type.is_filter
    ):
        return queryset

    filter_method = getattr(filters, "filter", None)
    if filter_method:
        return filter_method(queryset)

    filter_kwargs, filter_methods = build_filter_kwargs(filters)
    queryset = queryset.filter(**filter_kwargs)
    for filter_method in filter_methods:
        queryset = filter_method(queryset=queryset)
    return queryset


class StrawberryDjangoFieldFilters:
    def __init__(self, filters=UNSET, **kwargs):
        self.filters = filters
        super().__init__(**kwargs)

    @property
    def arguments(self) -> List[StrawberryArgument]:
        arguments = []
        if not self.base_resolver:
            filters = self.get_filters()
            if self.django_model and not self.is_list:
                if self.is_relation is False:
                    arguments.append(argument("pk", strawberry.ID))
            elif filters and filters is not UNSET:
                arguments.append(argument("filters", filters))
        return super().arguments + arguments

    def get_filters(self):
        if self.filters is not UNSET:
            return self.filters
        type_ = utils.unwrap_type(self.type or self.child.type)

        if utils.is_django_type(type_):
            return type_._django_type.filters
        return None

    def get_queryset(self, queryset, info, pk=UNSET, filters=UNSET, **kwargs):
        queryset = super().get_queryset(queryset, info, **kwargs)
        return apply(filters, queryset, pk)



from typing import Any

import strawberry
from django.contrib import auth
from django.contrib.auth.password_validation import validate_password

from strawberry_django.mutations.fields import (
    DjangoCreateMutation,
    get_input_data,
    update_m2m,
)

from ..resolvers import django_resolver


@django_resolver
def resolve_login(info, username: str, password: str):
    request = info.context.request
    user = auth.authenticate(request, username=username, password=password)
    if user is not None:
        auth.login(request, user)
        return user
    auth.logout(request)
    return None


@django_resolver
def resolve_logout(info) -> bool:
    request = info.context.request
    ret = request.user.is_authenticated
    auth.logout(request)
    return ret


def login() -> Any:
    mutation = strawberry.mutation(resolver=resolve_login)
    mutation.is_optional = True
    return mutation


def logout() -> bool:
    return strawberry.mutation(resolver=resolve_logout)


def register(user_type) -> Any:
    return DjangoRegisterMutation(user_type)


class DjangoRegisterMutation(DjangoCreateMutation):
    def create(self, data):
        input_data = get_input_data(self.input_type, data)
        validate_password(input_data["password"])
        instance = self.django_model.objects.create_user(**input_data)
        update_m2m([instance], data)
        return instance



import dataclasses
from inspect import cleandoc
from typing import Any, Dict, Generic, Optional, Type, TypeVar

import django
import django.db.models
import strawberry
from strawberry import UNSET
from strawberry.annotation import StrawberryAnnotation

from . import utils
from .fields.field import StrawberryDjangoField
from .fields.types import (
    get_model_field,
    is_optional,
    resolve_model_field_name,
    resolve_model_field_type,
)
from .settings import strawberry_django_settings as django_settings


_type = type


def get_type_attr(type_, field_name: str):
    attr = getattr(type_, field_name, UNSET)
    if attr is UNSET:
        attr = getattr(type_, "__dataclass_fields__", {}).get(field_name, UNSET)
    return attr


def get_field(
    django_type: "StrawberryDjangoType",
    field_name: str,
    field_annotation: Optional[StrawberryAnnotation] = None,
):
    if field_annotation is None:
        field_annotation = StrawberryAnnotation(None)
    attr = get_type_attr(django_type.origin, field_name)

    if utils.is_field(attr):
        field = django_type.field_cls.from_field(attr, django_type)
    else:
        field = django_type.field_cls(
            default=attr,
            type_annotation=field_annotation,
        )

    field.python_name = field_name
    if field_name in django_type.origin.__dict__.get("__annotations__", {}):
        # store origin django type for further usage
        field.origin_django_type = django_type

    if field_annotation:
        # annotation of field is used as a class type
        field.type_annotation = field_annotation
        field.is_auto = utils.is_auto(field_annotation)

    try:
        # resolve the django_name and check if it is relation field. django_name
        # is used to access the field data in resolvers
        django_name = field.django_name or field_name
        model_field = get_model_field(django_type.model, django_name)
        field.django_name = resolve_model_field_name(
            model_field, django_type.is_input, django_type.is_filter
        )
        field.is_relation = model_field.is_relation

        # Use the Django field help_text if no other description is available.
        settings = django_settings()
        if not field.description and settings["FIELD_DESCRIPTION_FROM_HELP_TEXT"]:
            model_field_help_text = getattr(model_field, "help_text", "")
            field.description = str(model_field_help_text) or None
    except django.core.exceptions.FieldDoesNotExist:
        if field.django_name or field.is_auto:
            raise  # field should exist, reraise caught exception
        model_field = None

    if field.is_relation:
        # change relation field type to auto if field is inherited from another
        # type. for example if field is inherited from output type but we are
        # configuring field for input type
        if not utils.is_similar_django_type(django_type, field.origin_django_type):
            field.is_auto = True

    if field.is_auto:
        # resolve type of auto field
        field_type = resolve_model_field_type(model_field, django_type)
        field.type_annotation = StrawberryAnnotation(field_type)

    if is_optional(model_field, django_type.is_input, django_type.is_partial):
        field.type_annotation.annotation = Optional[field.type_annotation.annotation]

    if django_type.is_input:
        if field.default is dataclasses.MISSING:
            # strawberry converts UNSET value to MISSING, let's set
            # it back to UNSET. this is important especially for partial
            # input types
            # TODO: could strawberry support UNSET default value?
            field.default_value = UNSET
            field.default = UNSET

    return field


def get_fields(django_type: "StrawberryDjangoType"):
    annotations = utils.get_annotations(django_type.origin)
    fields: Dict[str, StrawberryDjangoField] = {}

    # collect all annotated fields
    for field_name, field_annotation in annotations.items():
        field = get_field(django_type, field_name, field_annotation)
        fields[field_name] = field

    # collect non-annotated strawberry fields
    for field_name in dir(django_type.origin):
        if field_name in fields:
            continue
        attr = getattr(django_type.origin, field_name)
        if not utils.is_strawberry_field(attr):
            continue
        field = get_field(django_type, field_name)
        fields[field_name] = field

    return list(fields.values())


_O = TypeVar("_O", bound=type)
_M = TypeVar("_M", bound=django.db.models.Model)


@dataclasses.dataclass
class StrawberryDjangoType(Generic[_O, _M]):
    origin: _O
    model: Type[_M]
    is_input: bool
    is_partial: bool
    is_filter: bool
    filters: Any
    order: Any
    pagination: Any
    field_cls: Type[StrawberryDjangoField]


def process_type(
    cls,
    model: Type[django.db.models.Model],
    *,
    filters=UNSET,
    pagination=UNSET,
    order=UNSET,
    field_cls=UNSET,
    **kwargs
):
    original_annotations = cls.__dict__.get("__annotations__", {})

    if not field_cls or field_cls is UNSET:
        field_cls = StrawberryDjangoField

    django_type = StrawberryDjangoType(
        origin=cls,
        model=model,
        is_input=kwargs.get("is_input", False),
        is_partial=kwargs.pop("partial", False),
        is_filter=kwargs.pop("is_filter", False),
        filters=filters,
        order=order,
        pagination=pagination,
        field_cls=field_cls,
    )

    fields = get_fields(django_type)

    # update annotations and fields
    cls.__annotations__ = cls_annotations = {}
    for field in fields:
        annotation = (
            field.type
            if field.type_annotation is None
            else field.type_annotation.annotation
        )
        if annotation is None:
            annotation = StrawberryAnnotation(strawberry.auto)
        cls_annotations[field.name] = annotation
        setattr(cls, field.name, field)

    # Strawberry >= 0.92.0 defines `is_type_of` for types implementing
    # interfaces if the attribute has not been set yet. It allows only
    # instances of `cls` to be returned, we should allow model instances
    # too.
    if not hasattr(cls, "is_type_of"):
        cls.is_type_of = lambda obj, _info: isinstance(obj, (cls, model))

    # Get type description from either kwargs, or the model's docstring
    settings = django_settings()
    description = kwargs.pop("description", None)
    if not description and settings["TYPE_DESCRIPTION_FROM_MODEL_DOCSTRING"]:
        description = cleandoc(model.__doc__) or None

    strawberry.type(cls, description=description, **kwargs)

    # restore original annotations for further use
    cls.__annotations__ = original_annotations
    cls._django_type = django_type

    return cls


def type(model, *, filters=UNSET, **kwargs):
    def wrapper(cls):
        return process_type(cls, model, filters=filters, **kwargs)

    return wrapper


def input(model, *, partial=False, **kwargs):
    return type(model, partial=partial, is_input=True, **kwargs)


def mutation(model, **kwargs):
    return type(model, **kwargs)



import enum
from typing import List, Optional

import strawberry
from strawberry import UNSET
from strawberry.arguments import StrawberryArgument

import strawberry_django

from . import utils
from .arguments import argument


@strawberry.enum
class Ordering(enum.Enum):
    ASC = "ASC"
    DESC = "DESC"


def generate_order_args(order, prefix=""):
    args = []
    for field in strawberry_django.fields(order):
        ordering = getattr(order, field.name, UNSET)
        if ordering is UNSET:
            continue
        if ordering == Ordering.ASC:
            args.append(f"{prefix}{field.name}")
        elif ordering == Ordering.DESC:
            args.append(f"-{prefix}{field.name}")
        else:
            subargs = generate_order_args(ordering, prefix=f"{prefix}{field.name}__")
            args.extend(subargs)
    return args


def order(model):
    def wrapper(cls):
        for name, type_ in cls.__annotations__.items():
            if strawberry_django.is_auto(type_):
                type_ = Ordering
            cls.__annotations__[name] = Optional[type_]
            setattr(cls, name, UNSET)
        return strawberry.input(cls)

    return wrapper


def apply(order, queryset):
    if order is UNSET or order is None:
        return queryset
    args = generate_order_args(order)
    if not args:
        return queryset
    return queryset.order_by(*args)


class StrawberryDjangoFieldOrdering:
    def __init__(self, order=UNSET, **kwargs):
        self.order = order
        super().__init__(**kwargs)

    @property
    def arguments(self) -> List[StrawberryArgument]:
        arguments = []
        if not self.base_resolver:
            order = self.get_order()
            if order and order is not UNSET:
                arguments.append(argument("order", order))
        return super().arguments + arguments

    def get_order(self):
        if self.order is not UNSET:
            return self.order
        type_ = utils.unwrap_type(self.type or self.child.type)

        if utils.is_django_type(type_):
            return type_._django_type.order
        return None

    def get_queryset(self, queryset, info, order=UNSET, **kwargs):
        queryset = super().get_queryset(queryset, info, **kwargs)
        return apply(order, queryset)



from typing import Any

from strawberry import UNSET

from .fields import DjangoCreateMutation, DjangoDeleteMutation, DjangoUpdateMutation


def create(input_type=UNSET, permission_classes=[], **kwargs) -> Any:
    return DjangoCreateMutation(
        input_type, permission_classes=permission_classes, **kwargs
    )


def update(input_type=UNSET, filters=UNSET, permission_classes=[], **kwargs) -> Any:
    return DjangoUpdateMutation(
        input_type, filters=filters, permission_classes=permission_classes, **kwargs
    )


def delete(filters=UNSET, permission_classes=[], **kwargs) -> Any:
    return DjangoDeleteMutation(
        input_type=None,
        filters=filters,
        permission_classes=permission_classes,
        **kwargs
    )



from typing import List

import strawberry
from strawberry import UNSET
from strawberry.arguments import StrawberryArgument

from . import utils
from .arguments import argument


@strawberry.input
class OffsetPaginationInput:
    offset: int = 0
    limit: int = -1


def apply(pagination, queryset):
    if pagination is UNSET or pagination is None:
        return queryset

    start = pagination.offset
    stop = start + pagination.limit
    return queryset[start:stop]


class StrawberryDjangoPagination:
    def __init__(self, pagination=UNSET, **kwargs):
        self.pagination = pagination
        super().__init__(**kwargs)

    @property
    def arguments(self) -> List[StrawberryArgument]:
        arguments = []
        if not self.base_resolver:
            pagination = self.get_pagination()
            if pagination and pagination is not UNSET:
                arguments.append(argument("pagination", OffsetPaginationInput))
        return super().arguments + arguments

    def get_pagination(self):
        if self.pagination is not UNSET:
            return self.pagination

        type_ = utils.unwrap_type(self.type or self.child.type)
        if utils.is_django_type(type_):
            return type_._django_type.pagination
        return None

    def get_queryset(self, queryset, info, pagination=UNSET, **kwargs):
        queryset = super().get_queryset(queryset, info, **kwargs)
        return apply(pagination, queryset)



from typing import Optional

from strawberry import UNSET
from strawberry.annotation import StrawberryAnnotation
from strawberry.arguments import StrawberryArgument


def argument(name, type_, is_optional=True, default=UNSET):
    if is_optional:
        type_ = Optional[type_]
    return StrawberryArgument(
        default=default,
        description=None,
        graphql_name=None,
        python_name=name,
        type_annotation=StrawberryAnnotation(type_),
    )



from functools import _make_key
from typing import Callable, Dict, Hashable, Optional, Tuple

from django.core.cache import caches
from django.core.cache.backends.base import DEFAULT_TIMEOUT
from strawberry.extensions.base_extension import Extension


class DjangoCacheBase(Extension):
    """
    Base for a Cache that uses Django built in cache instead of an in memory cache

    Arguments:

    `cache_name: str`
        Name of the Django Cache to use, defaults to 'default'

    `timeout: Optional[float]`
        How long to hold items in the cache. See the Django Cache docs for details
        https://docs.djangoproject.com/en/4.0/topics/cache/

    `hash_fn: Optional[Callable[[Tuple, Dict], str]]`
        A function to use to generate the cache keys
        Defaults to the same key generator as functools.lru_cache
        WARNING! The default function does NOT work with memcached
        and will generate warnings
    """

    def __init__(
        self,
        cache_name: str = "default",
        timeout: Optional[float] = None,
        hash_fn: Optional[Callable[[Tuple, Dict], Hashable]] = None,
    ):
        self.cache = caches[cache_name]
        self.timeout = timeout or DEFAULT_TIMEOUT
        # Use same key generating function as functools.lru_cache as default
        self.hash_fn = hash_fn or (lambda args, kwargs: _make_key(args, kwargs, False))

    def execute_cached(self, func, *args, **kwargs):
        hash_key = self.hash_fn(args, kwargs)
        cache_result = self.cache.get(hash_key)
        if cache_result is not None:
            return cache_result
        func_result = func(*args, **kwargs)
        self.cache.set(hash_key, func_result, timeout=self.timeout)
        return func_result



def add_hook(field, hook_name, hooks):
    if hooks is None:
        hooks = []
    if not isinstance(hooks, (list, tuple)):
        hooks = [hooks]
    field._hooks[hook_name] = hooks

    def register_hook(func):
        field._hooks[hook_name].append(func)

    setattr(field, hook_name, register_hook)


def add(**kwargs):
    def wrapper(field):
        field._hooks = {}
        for hook_name, hooks in kwargs.items():
            add_hook(field, hook_name, hooks)

        def call_hooks(hook_name, caller):
            for hook in field._hooks[hook_name]:
                caller(hook)

        field._call_hooks = call_hooks
        return field

    return wrapper
